EasingButtons > Button {
    width: 100%;
}
EasingButtons {
    dock: left;
    overflow-y: scroll;
    width: 20;
}

#bar-container {
    content-align: center middle;
}

#duration-input {
    width: 30;
    background: $boost;
    padding: 0 1;
    border: tall transparent;
}

#duration-input:focus {
    border: tall $accent;
}

#inputs {
    padding: 1;
    height: auto;
    dock: top;
    background: $boost;
}

Bar {
    width: 1fr;
}

#other {
    width: 1fr;
    background: $panel;
    padding: 1;
    height: 100%;
    border-left: vkey $background;
}

#opacity-widget {
    padding: 1;
    background: $warning;
    color: $text;
    border: wide $background;
}

#label {
    width: auto;
    padding: 1;
}
